@@ -4,6 +4,7 @@ import logging
 from typing import *
 
 import torch
+import torchvision
 import PIL
 import numpy as np
 from bigotis.modeling_utils import ImageGenerator
@@ -226,11 +227,11 @@ class Aphantasia(ImageGenerator):
         self,
         prompt: str,
         lr: float = 3e-1,
-        img_save_freq: int = 1,
-        num_generations: int = 200,
+        num_steps: int = 200,
         num_random_crops: int = 20,
         height: int = 256,
         width: int = 256,
+        generation_cb=None,
     ) -> Tuple[List[PIL.Image.Image], List[torch.Tensor]]:
         """
         Returns a list of images generated while optimizing
@@ -239,16 +240,19 @@ class Aphantasia(ImageGenerator):
         Args:
             prompt (str): input prompt.
             lr (float, optional): learning rate. Defaults to 0.5.
-            img_save_freq (int, optional): rate at which a generation
-                is added to the final lis. Defaults to 1.
-            num_generations (int, optional): number of optimization 
+            num_steps (int, optional): number of optimization 
                 steps. Defaults to 200.
             num_random_crops (int, optional): number of augmented 
                 images to use. Defaults to 20.
             height (int, optional): height of the optimized image. 
                 Defaults to 256.
-            wifth (int, optional): wifth of the optimized image. 
+            width (int, optional): wifth of the optimized image. 
                 Defaults to 256.
+            generation_cb (function, optional): if provided, the 
+                function is executed for every optimization step. The
+                inputs of this function will be the generated image
+                from the VQGAN (Nx3xHxW) and the logits used to 
+                generate it (Nx256xDxD) and the optimization step.
 
         Returns:
             Tuple[List[PIL.Image.Image], List[torch.Tensor]]: list 
@@ -290,7 +294,7 @@ class Aphantasia(ImageGenerator):
 
         gen_img_list = []
         gen_fft_list = []
-        for step in range(num_generations):
+        for step in range(num_steps):
             loss = 0
 
             initial_img = self.fft_to_rgb(
@@ -311,34 +315,43 @@ class Aphantasia(ImageGenerator):
             )
             loss += 10 * self.compute_clip_loss(x_rec_stacked, prompt)
 
-            logging.info(f"\nIteration {step} of {num_generations}")
+            logging.info(f"\nIteration {step} of {num_steps}")
             logging.info(f"Loss {round(float(loss.data), 2)}")
 
             optimizer.zero_grad()
             loss.backward()
             optimizer.step()
 
-            if step % img_save_freq == 0:
-                with torch.no_grad():
-                    img = self.fft_to_rgb(
-                        fft_img=fft_img,
-                        scale=scale,
-                        height=height,
-                        width=width,
-                        shift=shift,
-                        contrast=1.0,
-                        decorrelate=True,
+            with torch.no_grad():
+                x_rec = self.fft_to_rgb(
+                    fft_img=fft_img,
+                    scale=scale,
+                    height=height,
+                    width=width,
+                    shift=shift,
+                    contrast=1.0,
+                    decorrelate=True,
+                )
+
+                # img = img.detach().cpu().numpy()[0]
+                # img = np.transpose(np.array(img)[:, :, :], (1, 2, 0))
+                # img = np.clip(img * 255, 0, 255).astype(np.uint8)
+
+                # pil_img = PIL.Image.fromarray(img)
+
+                x_rec_img = torchvision.transforms.ToPILImage(mode='RGB')(
+                    x_rec[0], )
+
+                gen_img_list.append(x_rec_img)
+                gen_fft_list.append(fft_img)
+
+                if generation_cb is not None:
+                    generation_cb(
+                        x_rec,
+                        fft_img,
+                        step,
                     )
 
-                    img = img.detach().cpu().numpy()[0]
-                    img = np.transpose(np.array(img)[:, :, :], (1, 2, 0))
-                    img = np.clip(img * 255, 0, 255).astype(np.uint8)
-
-                    pil_img = PIL.Image.fromarray(img)
-
-                    gen_img_list.append(pil_img)
-                    gen_fft_list.append(fft_img)
-
             torch.cuda.empty_cache()
 
         return gen_img_list, gen_fft_list
@@ -419,28 +432,34 @@ class Aphantasia(ImageGenerator):
 if __name__ == '__main__':
     prompt = "'Tempus Fugit' by Salvador Dali"
     lr = 0.5
-    img_save_freq = 1
-    num_generations = 64
+    num_steps = 64
     num_random_crops = 32
     height = 256
     width = 512
 
+    def save_img_cb(
+        x_rec,
+        _z_logits,
+        step,
+    ):
+        x_rec_img = torchvision.transforms.ToPILImage(mode='RGB')(x_rec[0])
+        x_rec_img.save(f"{step}.jpg")
+
     aphantasia = Aphantasia()
     gen_img_list, fft_logits_list = aphantasia.generate_from_prompt(
         prompt=prompt,
         lr=lr,
-        img_save_freq=img_save_freq,
-        num_generations=num_generations,
+        num_steps=num_steps,
         num_random_crops=num_random_crops,
         height=height,
         width=width,
+        generation_cb=save_img_cb,
     )
 
     _gen_img_list, fft_logits_list_ = aphantasia.generate_from_prompt(
         prompt="'Tempus Fugit', by Escher",
         lr=lr,
-        img_save_freq=img_save_freq,
-        num_generations=num_generations,
+        num_steps=num_steps,
         num_random_crops=num_random_crops,
         height=height,
         width=width,
