@@ -8,6 +8,8 @@ from typing import *
 import torch
 import torchvision
 import numpy as np
+import omegaconf
+import PIL
 from PIL import Image
 from omegaconf import OmegaConf
 from forks.taming_transformers.taming.models.vqgan import VQModel
@@ -17,15 +19,28 @@ logging.basicConfig(format='%(message)s', level=logging.INFO)
 
 
 class TamingDecoder(ImageGenerator):
+    """
+    Image generator that leverages the decoder of VQGAN to 
+    optimize images. This code uses the original implementation
+    from https://github.com/CompVis/taming-transformers.
+    """
     def __init__(
         self,
-        target_img_size=256,
-        device=None,
-    ):
+        device: str = 'cuda',
+    ) -> None:
+        """
+        Downloads the VQGAN model and loads a pre-defined 
+        configuration.
+
+        Args:
+            device (str, optional): defaults to 'cuda'.
+        """
         super().__init__()
 
-        self.target_img_size = target_img_size
-        self.embed_size = self.target_img_size // 16
+        self.supported_loss_types = [
+            "cosine_similarity",
+            "spherical_distance",
+        ]
 
         if device is not None:
             self.device = device
@@ -68,29 +83,22 @@ class TamingDecoder(ImageGenerator):
             ckpt_path=modeling_ckpt_path,
         ).to(self.device)
 
-    def vqgan_preprocess(
-        self,
-        img,
-    ):
-        img = img.convert("RGB")
-
-        min_img_dim = min(img.size)
-
-        scale_factor = self.target_img_size / min_img_dim
-        scaled_img_dim = (round(scale_factor * img.size[0]),
-                          round(scale_factor * img.size[1]))
-        img = img.resize(scaled_img_dim, Image.LANCZOS)
-        img = torchvision.transforms.functional.center_crop(
-            img, output_size=2 * [self.target_img_size])
-        img_tensor = torch.unsqueeze(torchvision.transforms.ToTensor()(img),
-                                     0).to(self.device)
-
-        img_tensor = 2. * img_tensor - 1.
-
-        return img_tensor
-
     @staticmethod
-    def load_config(config_path, display=False):
+    def load_config(
+        config_path: str,
+        display=False,
+    ) -> omegaconf.dictconfig.DictConfig:
+        """
+        Loads a VQGAN configuration file from a path or URL.
+
+        Args:
+            config_path (str): local path or URL of the config file.
+            display (bool, optional): if `True` the configuration is 
+                printed. Defaults to False.
+
+        Returns:
+            omegaconf.dictconfig.DictConfig: configuration dictionary.
+        """
         config = OmegaConf.load(config_path)
 
         if display:
@@ -100,9 +108,22 @@ class TamingDecoder(ImageGenerator):
 
     @staticmethod
     def load_vqgan(
-        config,
-        ckpt_path=None,
-    ):
+        config: omegaconf.dictconfig.DictConfig,
+        ckpt_path: str = None,
+    ) -> VQModel:
+        """
+        Load a VQGAN model from a config file and a ckpt path 
+        where a VQGAN model is saved.
+
+        Args:
+            config ([type]): VQGAN model config.
+            ckpt_path ([type], optional): path of a saved model. 
+                Defaults to None.
+
+        Returns:
+            VQModel: 
+                loaded VQGAN model.
+        """
         model = VQModel(**config.model.params)
 
         if ckpt_path is not None:
@@ -116,21 +137,64 @@ class TamingDecoder(ImageGenerator):
         self,
         prompt: str,
         lr: float = 0.5,
-        img_save_freq: int = 1,
-        num_generations: int = 200,
-        num_random_crops: int = 20,
-        init_img_path=None,
+        target_img_size: int = 256,
+        num_steps: int = 200,
+        num_augmentations: int = 32,
+        init_img_path: str = None,
         loss_type: str = 'cosine_similarity',
         loss_clip_value: float = None,
+        generation_cb=None,
         **kwargs,
-    ):
+    ) -> Tuple[List[PIL.Image.Image], List[torch.Tensor]]:
+        """
+        Returns a list of images generated while optimizing the
+        input of a pre-trained VQGAN decoder with a prompt.
+
+        Args:
+            prompt (str): input prompt.
+            lr (float, optional): learning rate. Defaults to 0.5.
+            target_img_size (int, optional): image size of the 
+                generated images. Defaults to 256.
+            num_steps (int, optional): number of optimization 
+                steps. Defaults to 200.
+            num_augmentations (int, optional): number of augmented 
+                images to use. Defaults to 20.
+            init_img_path (str, optional): path for an image to use
+                as the starting point of the optimization. Defaults 
+                to None.
+            loss_type (str, optional): either 'cosine_similarity' 
+                or 'spherical_distance'. Defaults to 
+                'cosine_similarity'.
+            loss_clip_value (float, optional): thresholding value for
+                for the CLIP embeddings. Defaults to None.
+            generation_cb (function, optional): if provided, the 
+                function is executed for every optimization step. The
+                inputs of this function will be the generated image
+                from the VQGAN (Nx3xHxW) and the logits used to 
+                generate it (Nx256xDxD) and the optimization step.
+
+        Returns:
+            Tuple[List[PIL.Image.Image], List[torch.Tensor]]: list 
+                of optimized images and their respective logits.
+        """
+        # TODO: implement batching
         batch_size = 1
 
+        target_img_size = target_img_size
+        embed_size = target_img_size // 16
+
+        if loss_type not in self.supported_loss_types:
+            print(
+                (f"ERROR! Loss type {loss_type} not recognized. "
+                 f"Only {' or '.join(self.supported_loss_types)} supported."))
+
+            return
+
         z_logits = .5 * torch.randn(
             batch_size,
             256,
-            self.embed_size,
-            self.embed_size,
+            embed_size,
+            embed_size,
         ).to(self.device)
 
         if init_img_path is None:
@@ -142,17 +206,19 @@ class TamingDecoder(ImageGenerator):
 
             init_img = torch.tensor(np.asarray(pil_img)).to(
                 self.device).float()[None, :]
-            init_img /= 255.
+            init_img = (init_img / 255.) * 2 - 1
             init_img = init_img.permute(0, 3, 1, 2)
 
             # clip_img_z_logits = self.get_clip_img_encodings(init_img)
             # clip_img_z_logits = clip_img_z_logits.detach().clone()
 
-            z, _, [_, _, indices] = self.vqgan_model.encode(init_img)
+            z, _, [_, _, _indices] = self.vqgan_model.encode(init_img)
 
             z_logits = torch.nn.Parameter(z)
             # img_z_logits = z.detach().clone()
 
+        torch.nn.Parameter(torch.ones((4, )))
+
         optimizer = torch.optim.AdamW(
             params=[z_logits],
             lr=lr,
@@ -162,20 +228,19 @@ class TamingDecoder(ImageGenerator):
 
         gen_img_list = []
         z_logits_list = []
-        for step in range(num_generations):
+        for step in range(num_steps):
             loss = 0
 
-            if step % img_save_freq == 0:
-                z_logits_list.append(z_logits.detach().clone())
+            z_logits_list.append(z_logits.detach().clone())
 
             z = self.vqgan_model.post_quant_conv(z_logits)
             x_rec = self.vqgan_model.decoder(z)
             x_rec = (x_rec.clip(-1, 1) + 1) / 2
             x_rec_stacked = self.augment(
                 x_rec,
-                num_crops=num_random_crops,
-                target_img_width=self.target_img_size,
-                target_img_height=self.target_img_size,
+                num_crops=num_augmentations,
+                target_img_width=target_img_size,
+                target_img_height=target_img_size,
             )
 
             loss += 10 * self.compute_clip_loss(
@@ -192,17 +257,22 @@ class TamingDecoder(ImageGenerator):
             #     loss += -10 * torch.cosine_similarity(
             #         self.get_clip_img_encodings(x_rec), clip_img_z_logits).mean()
 
-            logging.info(f"\nIteration {step} of {num_generations}")
+            logging.info(f"\nIteration {step} of {num_steps}")
             logging.info(f"Loss {round(float(loss.data), 2)}")
 
             optimizer.zero_grad()
             loss.backward()
             optimizer.step()
 
-            if step % img_save_freq == 0:
-                x_rec_img = torchvision.transforms.ToPILImage(mode='RGB')(
-                    x_rec[0])
-                gen_img_list.append(x_rec_img)
+            x_rec_img = torchvision.transforms.ToPILImage(mode='RGB')(x_rec[0])
+            gen_img_list.append(x_rec_img)
+
+            if generation_cb is not None:
+                generation_cb(
+                    x_rec,
+                    z_logits,
+                    step,
+                )
 
             torch.cuda.empty_cache()
 
@@ -210,10 +280,26 @@ class TamingDecoder(ImageGenerator):
 
     def interpolate(
         self,
-        z_logits_list,
-        duration_list,
+        z_logits_list: List[torch.Tensor],
+        duration_list: List[float],
         **kwargs,
-    ):
+    ) -> List[PIL.Image.Image]:
+        """
+        Returns a list of PIL images corresponding to the 
+        generations produced by interpolating the values
+        from `z_logits_list`.
+
+        Args:
+            z_logits_list (List[torch.Tensor]): list of VQVAE 
+                intermediate embeddings.
+            duration_list (List[float]): list with the duration of
+                the interpolation between each consecutive tensor. 
+                The last value represent the duration between the 
+                last tensor and the initial.
+
+        Returns:
+            List[PIL.Image.Image]: list of the resulting generated images.
+        """
         gen_img_list = []
         fps = 25
 
@@ -241,34 +327,42 @@ class TamingDecoder(ImageGenerator):
 
 
 if __name__ == '__main__':
-    prompt = "Landscape of Costa Rica"
+    target_img_size = 256
+    prompt = "A pink dog"
     lr = 0.5
-    img_save_freq = 1
-    num_generations = 32
-    num_random_crops = 16
-    # init_img_path = "alien.png"
-    init_img_path = None
+    num_steps = 200
+    num_augmentations = 128
     loss_type = 'cosine_similarity'
-    loss_clip_value = .1
+    init_img_path = None
+    # init_img_path = "alien.png"
+    loss_clip_value = None
+
+    def save_img_cb(
+        x_rec,
+        _z_logits,
+        step,
+    ):
+        x_rec_img = torchvision.transforms.ToPILImage(mode='RGB')(x_rec[0])
+        x_rec_img.save(f"{step}.jpg")
 
     taming_decoder = TamingDecoder()
     gen_img_list, z_logits_list = taming_decoder.generate_from_prompt(
         prompt=prompt,
         lr=lr,
-        img_save_freq=img_save_freq,
-        num_generations=num_generations,
-        num_random_crops=num_random_crops,
+        target_img_size=target_img_size,
+        num_steps=num_steps,
+        num_augmentations=num_augmentations,
         init_img_path=init_img_path,
         loss_type=loss_type,
         loss_clip_value=loss_clip_value,
+        generation_cb=save_img_cb,
     )
 
     # _gen_img_list, z_logits_list_ = taming_decoder.generate_from_prompt(
     #     prompt="Pokemon of type grass",
     #     lr=lr,
-    #     img_save_freq=img_save_freq,
-    #     num_generations=num_generations,
-    #     num_random_crops=num_random_crops,
+    #     num_steps=num_steps,
+    #     num_augmentations=num_augmentations,
     #     init_img_path=init_img_path,
     # )
 
